/**
  *  \file client/tiles/basetaskcmd.cc
  *  \brief WBaseAutoTaskCommandTile tile
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/basetaskcmd.h"
#include "client/dialogs/basetaskbuild.h"
#include "client/screens/planettaskscreen.h"
#include "client/tasksupport.h"
#include "client/widgets/buildorderdisplay.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/tables.h"
#include "int/args.h"
#include "int/if/baseif.h"
#include "int/if/basepred.h"
#include "int/if/ifutil.h"
#include "int/predict.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "util/nls.h"
#include "util/strtab.h"

static const char*const buttons[] = {
    N_("1 - Orders"),
    N_("2 - Cargo"),
    N_("3 - Misc."),
};

namespace {
    /** Starbase Auto Task Build Command Parser. This parses a single
        independant BuildShip/EnqueueShip command. */
    class WBaseTaskBuildCommandParser : public IntAutoTaskPredictor {
        TBuildOrder order;
        string_t verb;
     public:
        WBaseTaskBuildCommandParser();
        bool predictInstruction(const string_t& name, IntArgBlock& args);
        const TBuildOrder& getOrder() const;
        const string_t& getVerb() const;
    };
}

/********************** WBaseTaskBuildCommandParser **********************/

WBaseTaskBuildCommandParser::WBaseTaskBuildCommandParser()
{
    order.hull = 0;
}

bool
WBaseTaskBuildCommandParser::predictInstruction(const string_t& name, IntArgBlock& args)
{
    if (name == "BUILDSHIP" || name == "ENQUEUESHIP") {
        if (parseBuildShipCommand(args, order)) {
            verb = name;
        }
    }
    return true;
}

inline const TBuildOrder&
WBaseTaskBuildCommandParser::getOrder() const
{
    return order;
}

inline const string_t&
WBaseTaskBuildCommandParser::getVerb() const
{
    return verb;
}



/************************ WBaseAutoTaskCommandTile ***********************/

WBaseAutoTaskCommandTile::WBaseAutoTaskCommandTile(GObjectType& type)
    /* We actually need 5 lines; there's room for 7, so we take
       those for safety in case something wraps. */
    : WTile(GfxPoint(315, 7*font_heights[FONT_NORMAL])),
      attype(dynamic_cast<WPlanetAutoTaskType*>(&type))
{
    init();
}

WBaseAutoTaskCommandTile::~WBaseAutoTaskCommandTile()
{ }

void
WBaseAutoTaskCommandTile::init()
{
    // Observer
    if (attype) {
        conn_task_changed =
            attype->sig_task_changed.add(static_cast<UIBaseWidget*>(this),
                                         (void (UIBaseWidget::*)()) &UIBaseWidget::drawWidget);
    }

    // Controls
    setLayout(RightJustified, 0, true);
    StaticTranslatedStringTable tab(buttons, countof(buttons));
    const int buttonWidth = getTextWidth(*fonts[FONT_SMALL], tab) + 20;
    const int height = font_heights[FONT_SMALL] * 3/2;
    for (uint_t i = 0; i < countof(buttons); ++i) {
        UIButton& btn = h.add(new UIButton(1000+i, tab.get(i), '1'+i));
        btn.addCommand();
        btn.setState(UIButton::bf_LeftJust + UIButton::bf_SmallSize, true);
        addControl(GfxRect(0, (height+2)*i, buttonWidth, height), btn);
    }

    btn_edit = &h.add(new UIButton("E", 'e'));
    addControl(GfxRect(buttonWidth-20, 7*font_heights[FONT_NORMAL]-20, 20, 20), *btn_edit);
    
    widthReserve = 5 + buttonWidth;
}

void
WBaseAutoTaskCommandTile::drawData(GfxCanvas& can)
{
    // Sanity check
    if (attype == 0) {
        return;
    }

    // Clear
    GfxRect area = getInnerSize(5, widthReserve);
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, area);

    // Label
    ctx.setTextAlign(2, 1);
    ctx.setColor(UIColor::tc_Static);
    outText(ctx, getExtent().w - 25, area.h - 10, _("Edit"));
    ctx.setTextAlign(0, 0);

    // Parse current command
    WBaseTaskBuildCommandParser parser;
    parser.predictStatement(attype->getEditor(), attype->getCursor());

    if (parser.getOrder().hull != 0) {
        // It's a valid build order, display it
        ctx.setColor(UIColor::tc_Contrast);
        outTextF(ctx, area.x, area.y, area.w, _("Ship Build Order:"));

        ctx.setColor(UIColor::tc_Static);
        outTextFormatted(ctx, area.x, area.y+font_heights[FONT_NORMAL], area.w,
                         formatBuildOrder(parser.getOrder()));
        btn_edit->setState(UIButton::st_Disabled, false);
    } else {
        // Not valid. Disable button.
        btn_edit->setState(UIButton::st_Disabled, true);
    }
}

void
WBaseAutoTaskCommandTile::createSetMissionCommand()
{
    // Find out old mission
    GPlanet* pl = dynamic_cast<GPlanet*>(attype->getCurrentObject());
    if (!pl) {
        return;
    }
    WBaseTaskPredictor pred(*pl);
    pred.predictTask(attype->getEditor(), attype->getCursor());

    // FIXME: copied from doBaseChangeMission
    UIStringList data;
    for (int i = 0; i <= 6; ++i)
        data.add(i, format("%d - %s", i, getBaseMissionName(i)));

    UIStandardListbox lb(0, 7, 300, data);
    lb.scrollToKey(pred.getPlanet().getBaseMission());
    if (lb.doStandardListbox(_("Starbase Mission"), 0 /* FIXME: help id */)) {
        attype->insert(format("SetMission %d   %% %s", lb.getKey(), getBaseMissionName(lb.getKey())));
    }
}

void
WBaseAutoTaskCommandTile::createEnqueueShipCommand()
{
    // Find out preconditions
    GPlanet* pl = dynamic_cast<GPlanet*>(attype->getCurrentObject());
    if (!pl) {
        return;
    }
    WBaseTaskPredictor pred(*pl);
    pred.predictTask(attype->getEditor(), attype->getCursor());

    // We'll be making an EnqueueShip command, which will wait until its
    // preceding command is performed. So do that now.
    if (pred.getPlanet().getBaseBuildOrderHullSlot() > 0) {
        pred.advanceTurn();
    }

    // What can we build?
    std::vector<int> hulls;
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
        if (int h = getTruehull(pl->getOwner(), i)) {
            hulls.push_back(h);
        }
    }

    // Build dialog
    WBaseTaskShipBuildDialog dialog(pred.getPlanet(), hulls);
    ui_root->centerWidget(dialog);
    if (dialog.runDialog()) {
        attype->insert(makeBuildOrderCommand("EnqueueShip", dialog.getBuildOrder()));
    }
}

void
WBaseAutoTaskCommandTile::editCurrentCommand()
{
    // Are we looking at a build command?
    WBaseTaskBuildCommandParser parser;
    parser.predictStatement(attype->getEditor(), attype->getCursor());
    if (!parser.getOrder().hull) {
        return;
    }

    // Find out preconditions
    GPlanet* pl = dynamic_cast<GPlanet*>(attype->getCurrentObject());
    if (!pl) {
        return;
    }
    WBaseTaskPredictor pred(*pl);
    pred.predictTask(attype->getEditor(), attype->getCursor());

    // We'll be making an EnqueueShip command, which will wait until its
    // preceding command is performed. So do that now.
    if (pred.getPlanet().getBaseBuildOrderHullSlot() > 0) {
        pred.advanceTurn();
    }

    // What can we build?
    std::vector<int> hulls;
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
        if (int h = getTruehull(pl->getOwner(), i)) {
            hulls.push_back(h);
        }
    }
    pred.getPlanet().setBaseBuildOrder(parser.getOrder());

    // Build dialog
    WBaseTaskShipBuildDialog dialog(pred.getPlanet(), hulls);
    ui_root->centerWidget(dialog);
    if (dialog.runDialog()) {
        string_t cmd = parser.getVerb();
        if (cmd == "ENQUEUESHIP") {
            cmd = "EnqueueShip";
        }
        if (cmd == "BUILDSHIP") {
            cmd = "BuildShip";
        }
        attype->replace(attype->getCursor(), 1, makeBuildOrderCommand(cmd, dialog.getBuildOrder()));
    }
}

bool
WBaseAutoTaskCommandTile::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case 1000:
        if (attype != 0) {
            StringList sl;
            sl.add(1, _("Enqueue Ship Build Order"));
            sl.add(2, _("Set Mission"));
            sl.add(3, _("Set FCode"));
            UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
            if (list.doStandardMenu(sender)) {
                switch (list.getKey()) {
                 case 1:
                    createEnqueueShipCommand();
                    break;
                 case 2:
                    createSetMissionCommand();
                    break;
                 case 3:
                    attype->createSetFCodeCommand();
                    break;
                }
            }
        }
        return true;
     case 1001:
        if (attype != 0) {
            attype->createCargoCommand(sender);
        }
        return true;
     case 1002:
        if (attype != 0) {
            attype->createCommonCommand(sender, false);
        }
        return true;
    }
    return WTile::handleCommand(cmd, sender);
}

bool
WBaseAutoTaskCommandTile::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'e':
            if (attype != 0) {
                editCurrentCommand();
            }
            return true;
        }
    }
    return WTile::handleEvent(event, second_pass);
}
