/**
  *  \file game/inbox.cc
  *  \brief Subspace Messages: Inbox
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "game/inbox.h"
#include "game/pref.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/nls.h"
#include "game/racename.h"
#include "game/struct.h"
#include "game/player.h"

/********************************* GInbox ********************************/

/** Message structure. */
struct GInbox::Message {
    string_t text;
    int      receiver;

    Message(const string_t& str, int receiver)
        : text(str), receiver(receiver)
        { }
};

/** Constructur. Creates empty messagebox. */
GInbox::GInbox()
{ }

/** Destructor. */
GInbox::~GInbox()
{ }

/** Add a single message.
    \param str Complete text of message
    \param receiver Receiver of this message */
void
GInbox::addMessage(const string_t str, int receiver)
{
    messages.push_back(new Message(str, receiver));
}

int
GInbox::getCount()
{
    return messages.size();
}

string_t
GInbox::getText(int index)
{
    ASSERT(index >= 0 && unsigned(index) < messages.size());
    return messages[index]->text;
}

/** Simplify message header.
    - Remove surrounding angle brackets
    - Remove optional "(n)" produced by PHCc
    - Lowercase all-caps headings */
static string_t
tweakHeader(string_t what)
{
    string_t::size_type n;
    while ((n = what.find_first_of("<>")) != string_t::npos) {
        what.erase(n, 1);
    }
    while (what.size() && what[what.size()-1] == ' ') {
        what.erase(what.size()-1);
    }
    if (what.size() && what[what.size()-1] == ')') {
        // "PHCc game info (1)"
        n = what.rfind('(');
        if (n != string_t::npos && n > 4)
            what.erase(n);
    }
    return strTrim(strLCWords(what));
}

/** Get Message Heading. This is the same algorithm as in PCC 1.x. */
string_t
GInbox::getHeading(int index)
{
    using namespace std;
    string_t line = strTrim(strFirst(getText(index), "\n"));
    if (line.length() < 5)
        /* translators: must start with "(  )" */
        return _("(  ) Unknown");
    if (line[0] != '(')
        /* pre-3.2 message format */
        return "(  ) " + tweakHeader(line);
    if (line[1] == 'o' && getUserPreferences().CollapseOldMessages())
        /* translators: must start with "(oX)" */
        return _("(oX) Message from earlier turn");

    /* It is a message in our preferred format */
    string_t pre = line.substr(0, 3) + ") ";
    switch(toupper((unsigned char) pre[2])) {
     case 'R':
        if (line.size() > 3) {
            int i = getPlayerIdFromChar(line[3]);
            if (i > 0)
                return pre + player_racenames.getFullName(i);
            else if (i == 0)
                return pre + _("Anonymous Message");
            else
                break;
        }
     case 'D':
        return pre + _("Starbase Message");
     case 'L':
        return pre + _("Minefield Laid");
     case 'I':
        return pre + _("Ion Storm");
     case 'G':
        return pre + _("HConfig");
     case 'M':
        /* Mine scan/sweep. People want to separate these, to be able to filter
           out unsuccessful scans, and only see sweeps. We look for two keyphrases
           to detect sweeps. We only look into the message if we're reasonably
           sure that it is an English message to avoid false positives (hence the
           check for the header line).

           This is the same logic as used in PCC 1.1.16, and it works for English
           and NewEnglish (and, implicitly, for German, which happens to use
           distinct headers for scan and sweep). */
        if (line.find("Sub Space Message") != line.npos) {
            string_t fulltext = getText(index);
            if (fulltext.find("ines have been destroyed") != fulltext.npos
                || fulltext.find("is using beam weapons to") != fulltext.npos)
            {
                return pre + _("Mine Sweep");
            } else {
                return pre + _("Mine Scan");
            }
        }
        /* else use default */
        break;
    }
    strRemove(line, ")");
    return pre + tweakHeader(line);
}

/** Get message receiver. */
int
GInbox::getReceiver(int index) const
{
    ASSERT(index >= 0 && unsigned(index) < messages.size());
    return messages[index]->receiver;
}

/** Sort messages. This will group messages of equal subjects
    together, but preserves the overall order (i.e. messages from
    other races will remain first). */
void
GInbox::sort()
{
    ptr_vector<Message> new_data;
    std::vector<bool> did;
    did.resize(messages.size());
    new_data.reserve(messages.size());

    /* This algorithm is nominally O(n^2). */
    for (unsigned int i = 0; i < messages.size(); ++i) {
        if (!did[i]) {
            const string_t h = getHeading(i);

            // Copy over all messages with this subject
            for (unsigned int j = i; j < messages.size(); ++j) {
                if (!did[j] && getHeading(j) == h) {
                    did[j] = true;
                    new_data.push_back(messages.extract_elem(j));
                }
            }
        }
    }
    messages.swap(new_data);
}

/** Load Inbox. This reads RST and MDATA files.
    \param str File. Pointer is set to beginning of message directory.
    File must be seekable.
    \throws FileFormatException on error. Caller should add file name. */
void
GInbox::loadInbox(Stream& str, int receiver)
{
    /* read count */
    TInt16 count;
    if (!getStructure(str, count))
        throw FileFormatException(str, _("Message file too short"));
    if (count.value < 0)
        throw FileFormatException(str, _("Message file is invalid"));
    if (count.value == 0)
        return;

    /* read message directory */
    uint32 size = count.value * uint32(6);
    Array<char> dir(size);
    if (str.read(dir.ptr(), size) != size)
        throw FileFormatException(str, _("Message file too short"));

    int16 max_size = 0;
    for (int i = 0; i < count.value; ++i) {
        /* verify message directory entry */
        TIncomingMessageHeader hdr;
        getStructure(dir.ptr() + 6*i, hdr);
        if (hdr.address <= 0 || hdr.length <= 0)
            throw FileFormatException(str, _("Message file is invalid"));
        if (hdr.length > max_size)
            max_size = hdr.length;
    }

    /* load the messages */
    Array<char> buffer(max_size);
    for (int i = 0; i < count.value; ++i) {
        TIncomingMessageHeader hdr;
        getStructure(dir.ptr() + 6*i, hdr);
        str.seek(hdr.address-1);
        str.readT(buffer.ptr(), hdr.length);
        addMessage(getMessageFromArray(buffer.ptr(), hdr.length, true), receiver);
    }
}
