/**
  *  \file int/compound.cc
  *  \brief Interpreter: Compound Values
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/compound.h"
#include "int/interror.h"
#include "int/context.h"
#include "int/vmio.h"


bool
IntIndexableValue::isProcedureCall()
{
    return false;
}

IntValue*
IntIndexableValue::call(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    return get(args);
}


/************************ IntSimpleIndexableValue ************************/

/** Constructor.
    \param get  Function to implement get() (=indexed read access, function call).
                Can be null, in this case read access ('pushind' opcode) will not
                be permitted.
    \param make Function to implement makeFirstContext() (=iteration access).
                Can be null, in this case iteration ('firstindex' opcode) will not
                be permitted. */
IntSimpleIndexableValue::IntSimpleIndexableValue(get_t get, make_t make)
    : get_function(get), make_function(make)
{ }

IntSimpleIndexableValue::~IntSimpleIndexableValue()
{ }

// IntIndexableValue:
IntValue*
IntSimpleIndexableValue::get(IntArgBlock& args)
{
    if (get_function) {
        return get_function(args);
    } else {
        throw IntError::typeError(IntError::tyIndexable);
    }
}

void
IntSimpleIndexableValue::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntContext*
IntSimpleIndexableValue::makeFirstContext()
{
    if (make_function)
        return make_function();
    else
        throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
string_t
IntSimpleIndexableValue::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntSimpleIndexableValue::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntSimpleIndexableValue::clone()
{
    return new IntSimpleIndexableValue(*this);
}

/************************ IntSimpleProcedureValue ************************/

IntSimpleProcedureValue::IntSimpleProcedureValue(call_t call)
    : call_function(call)
{ }

IntSimpleProcedureValue::~IntSimpleProcedureValue()
{ }

// IntCallableValue:
IntValue*
IntSimpleProcedureValue::call(IntExecutionContext& exc, IntArgBlock& args)
{
    call_function(exc, args);
    return 0;
}

bool
IntSimpleProcedureValue::isProcedureCall()
{
    return true;
}

// IntValue:
string_t
IntSimpleProcedureValue::toString(bool /*readable*/)
{
    return "#<procedure>";
}

void
IntSimpleProcedureValue::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntSimpleProcedureValue::clone()
{
    return new IntSimpleProcedureValue(call_function);
}

/*************************** IntSubroutineValue **************************/

IntSubroutineValue::IntSubroutineValue(IntBCORef bco)
    : bco(bco)
{ }

IntSubroutineValue::~IntSubroutineValue()
{ }

IntBCORef
IntSubroutineValue::getBytecodeObject()
{
    return bco;
}

string_t
IntSubroutineValue::toString(bool /*readable*/)
{
    string_t name = bco->getName();
    string_t result = "#<subr";
    if (name.size()) {
        result += ":";
        result += name;
    }
    return result + ">";
}

void
IntSubroutineValue::store(IntTagNode& tag, Stream& /*aux*/)
{
    IntVMSaveContext* vsc = IntVMSaveContext::getCurrentInstance();
    if (vsc != 0) {
        tag.tag   = IntTagNode::Tag_BCO;
        tag.value = vsc->addBCO(*bco);
    } else {
        throw IntError::notSerializable();
    }
}

IntSubroutineValue*
IntSubroutineValue::clone()
{
    return new IntSubroutineValue(bco);
}
