/**
  *  \file int/if/conif.cc
  *  \brief Interpreter Game Interface - Console Components
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/compound.h"
#include "int/data.h"
#include "int/exec.h"
#include "int/if/conif.h"
#include "int/if/ifutil.h"
#include "int/intglobal.h"
#include "int/names.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/recode.h"
#include "int/args.h"
#include "arch/platform.h"
#include "cpluslib/utf8.h"
#include "util/console.h"

/** Check whether an Unicode character should be accepted. */
static bool
acceptUnicode(uint32_t uni, int32_t flags)
{
    /* This is almost a copy from UIInputLine::acceptUnicode */
    if (uni < 32 || uni == 127) {
        // Refuse controls
        return false;
    }
    if ((flags & 1) != 0 && (uni < '0' || uni > '9')) {
        // Refuse nondigits if requested
        return false;
    }
    if ((flags & 2) != 0 && (uni >= 128)) {
        // Refuse non-ASCII if requested
        return false;
    }
    if ((flags & 16) != 0 && !isReliableGameCharacter(uni)) {
        // Refuse non-game characters if requested
        return false;
    }
    return true;
}


/** Global command "UI.Input", console version. */
static void
IFUIInput(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.Input <prompt>[, <title>, <maxChars>, <flags>, <default>] */
    args.checkArgumentCount(1, 5);

    string_t prompt;
    string_t title;
    string_t defaultText;
    int32_t  maxChars = 255;
    int32_t  flags = 0;
    int32_t  width = 0;

    // Mandatory argument
    if (!checkStringArg(prompt, args.getNext()))
        return;

    // Optional arguments
    title = prompt;
    checkStringArg(title, args.getNext());
    checkIntArg(maxChars, args.getNext(), 0, 32000);
    checkFlagArg(flags, &width, args.getNext(), "NHPFGM");
    checkStringArg(defaultText, args.getNext());

    // Convert flags
    //   N = 1 = numeric
    //   H = 2 = on high ASCII
    //   P = 4 = password masking
    //   F = 8 = frame
    //   G = 16 = game charset
    //   M = 32 = width is in ems
    while (1) {
        // Show prompt and get initial input
        string_t line;
        if (!Platform::readConsoleLine(prompt + "> ", line)) {
            // EOF
            exc.setVariable("UI.RESULT", 0);
            return;
        }

        line = substrUtf8(line, 0, maxChars);

        // Check whether input is valid, by decoding and sanitizing the UTF-8.
        string_t result;
        bool hadInvalidChars = false;
        Utf8Reader rdr(line.data(), line.size());
        for (uint32_t i = 0; rdr.hasMore() && i < uint32_t(maxChars); ++i) {
            uint32_t ch = rdr.get();
            if (acceptUnicode(ch, flags)) {
                appendUtf8(result, ch);
            } else {
                hadInvalidChars = true;
                appendUtf8(result, (flags & 1) ? '0' : '?');
            }
        }
        if (!hadInvalidChars) {
            // Accepted input
            IntStringValue sv(line);
            exc.setVariable("UI.RESULT", &sv);
            return;
        }

        // Not accepted, continue
        console << _("Your input contains characters that are not permitted at this place.\nPlease try again.\n");
    }
}


static void
initInterpreterConsoleInterface()
{
    defineGlobalValue("SYSTEM.GUI", makeBoolValue(0));
    defineGlobalProcedure("UI.INPUT", IFUIInput);
    // MessageBox -> core.q
    // UI.Update -> core.q
}

// Constraints: after mod_intgame_common, after mod_int, with mod_base_console
const ModuleInitializer mod_intgame_console[] = {
    { OnStartup, initInterpreterConsoleInterface },
    { 0, 0 }
};
