/**
  *  \file int/if/planetif.cc
  *  \brief Interpreter Game Interface: Planets
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/planetif.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/planet.h"
#include "int/compound.h"
#include "int/data.h"
#include "int/if/baseif.h"
#include "int/if/ifutil.h"
#include "int/if/playerif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "int/value.h"
#include "game/tables.h"
#include "game/ship.h"
#include "game/team.h"
#include "cpluslib/format.h"
#include "int/if/objif.h"
#include "game/actions/planetbase.h"
#include "game/actions/planetstructs.h"
#include "int/exec.h"
#include "game/actions/planetsupplies.h"
#include "game/undo.h"
#include "int/args.h"
#include "game/cspec.h"
#include "int/contextio.h"
#include "int/if/shipif.h"
#include "int/process.h"
#include "int/propenum.h"
#include "game/cargotr.h"

/** Implementation of "Planet(nn)".
    \param args function parameters
    \return value representing "Planet(nn)" */
IntValue*
IFPlanetGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_PLANETS))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return new IntPlanetContext(id);
}

/** Implementation of "ForEach Planet".
    \return Context representing first planet */
IntContext*
IFPlanetMake()
{
    if (!haveDisplayedTurn())
        return 0;

    int id = global_anyplanet_type.getNextNoWrap(0, false);
    if (id == 0)
        return 0;
    else
        return new IntPlanetContext(id);
}

/***************************** Planet Methods ****************************/

static void
doStructureBuild(GPlanet& pl, IntExecutionContext& exc, IntValue* countVal, IntValue* flagVal, TPlanetaryStructure str)
{
    int32_t count, flag;
    if (!checkIntArg(count, countVal, -10000, +10000))
        return;

    flag = 0;
    checkFlagArg(flag, 0, flagVal, "N");

    GMultiTransaction mt;
    GPlanetTransfer& transfer         = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GPlanetBuildStructuresAction& bsa = mt.addNewSubtransaction(new GPlanetBuildStructuresAction(pl, transfer));
    bsa.setUndoInformation(getDisplayedTurn().getPreviousUniverse().getPlanet(pl.getId()));

    int built = bsa.addStructuresLimitCash(str, count);
    if (flag) {
        // We permit partial build, place remainder in BUILD.REMAINDER
        IntIntValue iv(count - built, false);
        exc.setVariable("BUILD.REMAINDER", &iv);
        mt.doCommit();
    } else {
        // We do not permit partial build, so refuse it
        if (built != count) {
            mt.doCancel();
            throw GError(GError::ePerm, GError::ePerm);
        } else {
            mt.doCommit();
        }
    }
}

static void
IFPlanetAutoBuild(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    GPlanet& pl = static_cast<GPlanet&>(obj);
    GMultiTransaction mt;
    GPlanetTransfer& transfer         = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GPlanetBuildStructuresAction& bsa = mt.addNewSubtransaction(new GPlanetBuildStructuresAction(pl, transfer));
    bsa.doStandardAutoBuild();
    mt.doCommit();
}

static void
IFPlanetBuildBase(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // Check arguments
    bool doit = true;
    args.checkArgumentCount(0, 1);
    if (args.getNumArgs() > 0)                // FIXME: this is inconsistent in that a missing arg means true, not ignore
        if (!checkBoolArg(doit, args.getNext()))
            return;

    // Do it
    GPlanet& pl = static_cast<GPlanet&>(obj);
    GMultiTransaction mt;
    GPlanetTransfer& transfer = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    mt.addNewSubtransaction(new GPlanetBuildStarbaseAction(pl, transfer, doit));
    mt.doCommit();
}

static void
IFPlanetBuildBaseDefense(GObject& obj, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);
    IntValue* amount = args.getNext();
    IntValue* flag = args.getNext();
    doStructureBuild(static_cast<GPlanet&>(obj),
                     exc,
                     amount,
                     flag,
                     str_BaseDefense);
}

static void
IFPlanetBuildDefense(GObject& obj, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);
    IntValue* amount = args.getNext();
    IntValue* flag = args.getNext();
    doStructureBuild(static_cast<GPlanet&>(obj),
                     exc,
                     amount,
                     flag,
                     str_Defense);
}

static void
IFPlanetBuildFactories(GObject& obj, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);
    IntValue* amount = args.getNext();
    IntValue* flag = args.getNext();
    doStructureBuild(static_cast<GPlanet&>(obj),
                     exc,
                     amount,
                     flag,
                     str_Factories);
}

static void
IFPlanetBuildMines(GObject& obj, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);
    IntValue* amount = args.getNext();
    IntValue* flag = args.getNext();
    doStructureBuild(static_cast<GPlanet&>(obj),
                     exc,
                     amount,
                     flag,
                     str_Mines);
}

static void
IFPlanetSellSupplies(GObject& obj, IntExecutionContext& exc, IntArgBlock& args)
{
    // Fetch arguments
    args.checkArgumentCount(1, 2);

    int32_t amount;
    if (!checkIntArg(amount, args.getNext()))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    // Get planet
    GPlanet& pl = static_cast<GPlanet&>(obj);

    // Do it
    int32_t undid;
    if (amount == 0) {
        // No change
        undid = 0;
    } else if (amount > 0) {
        // Sell supplies. We have a transaction for that.
        GPlanetSellSuppliesAction action(pl);
        action.setAmount(amount);
        undid = amount - action.getAmount();
        if (!flag && undid != 0) {
            action.doCancel();
            throw GError(GError::ePerm, GError::ePerm);
        }
        action.doCommit();
    } else {
        // Buy supples. We do not yet have a transaction for that. FIXME.
        if (!pl.isPlayed())
            throw GError(GError::eNotPlaying, GError::eNotPlaying);

        // Compute how much we can buy
        GUndoInformation undo;
        undo.computeInformation(getDisplayedTurn(), pl.getId());
        amount = -amount;

        int32 supplies = undo.getSuppliesAllowedToBuy();
        int32 money    = pl.getCargo(el_Money);
        int32 maximum  = supplies > money ? money : supplies;

        if (amount > maximum) {
            undid = maximum - amount;     // should be negative
            if (flag)
                amount = maximum;
            else
                throw GError(GError::ePerm, GError::ePerm);
        } else {
            undid = 0;
        }

        // Do it.
        pl.changeCargoRaw(el_Supplies, +amount);
        pl.changeCargoRaw(el_Money,    -amount);
    }

    // Set variable
    if (flag) {
        IntIntValue iv(undid, false);
        exc.setVariable("BUILD.REMAINDER", &iv);
    }
}

static void
IFPlanetSetColonistTax(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setPlanetProperty(static_cast<GPlanet&>(obj), ippColonistTax, args.getNext());
}

static void
IFPlanetSetComment(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    if (IntValue* v = args.getNext()) {
        IntStringValue sv(v->toString(false));
        int sid = obj.getId();
        getPlanetProperties(sid).set(sp_Comment, &sv);
        global_anyplanet_type.getObject(sid).markDirty();
    }
}

static void
IFPlanetSetFCode(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setPlanetProperty(static_cast<GPlanet&>(obj), ippFCode, args.getNext());
}

static void
IFPlanetSetNativeTax(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setPlanetProperty(static_cast<GPlanet&>(obj), ippNativeTax, args.getNext());
}

static void
IFPlanetCargoTransfer(GObject& obj, IntExecutionContext& exc, IntArgBlock& args)
{
    // Args are cargospec, ship Id, OSNJ flags + proxy ID
    // FIXME: proxy
    string_t cargospec;
    int32_t target_sid;
    int32_t flags = 0;
    args.checkArgumentCount(2, 3);
    if (!checkStringArg(cargospec, args.getNext()))
        return;
    if (!checkIntArg(target_sid, args.getNext(), 1, NUM_SHIPS))
        return;
    checkFlagArg(flags, 0, args.getNext(), "OSN");

    // Validate
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GPlanet& pl     = static_cast<GPlanet&>(obj);
    GShip& target   = univ.getShip(target_sid);
    if (!pl.isExistant() || !target.isExistant()) {
        throw IntError("Ship does not exist");
    }
    if (target.getPos() != pl.getPos()) {
        throw IntError(GError::ePos);
    }
    if (target.getKind() != GShip::CurrentShip) {
        throw IntError(GError::eNotPlaying);
    }

    // Build the transfer structures
    GCargoTransfer tx;
    tx.addNewContainer(makePlanetTransfer(pl.getId(), target));
    tx.addNewContainer(new GShipTransfer(target));
    doScriptTransfer(exc, tx, cargospec, 0, flags);
    tx.doCommit();
}

static void
IFPlanetAutoTaxColonists(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    GPlanet& pl = static_cast<GPlanet&>(obj);
    if (!pl.isPlayed()) {
        throw IntError(GError::eNotPlaying);
    }

    mp16_t val = pl.getColonistSafeTax(pl.getStructures(str_Mines) + pl.getStructures(str_Factories));
    if (val.isKnown()) {
        pl.setColonistTax(val);
    }
}

static void
IFPlanetAutoTaxNatives(GObject& obj, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    GPlanet& pl = static_cast<GPlanet&>(obj);
    if (!pl.isPlayed()) {
        throw IntError(GError::eNotPlaying);
    }
    mp16_t val = pl.getNativeSafeTax(pl.getStructures(str_Mines) + pl.getStructures(str_Factories));
    if (val.isKnown()) {
        pl.setNativeTax(val);
    }
}

/**************************** IntPlanetContext ***************************/

namespace { enum { PlanetProperty, BaseProperty, OwnerProperty, PlanetMethod }; }

static const TIntMapping planet_mapping[] = {
    { "AUTOBUILD",                 ipmAutoBuild,         PlanetMethod,   thProcedure },
    { "AUTOTAXCOLONISTS",          ipmAutoTaxColonists,  PlanetMethod,   thProcedure },
    { "AUTOTAXNATIVES",            ipmAutoTaxNatives,    PlanetMethod,   thProcedure },
    { "BASE",                      ippBaseStr,           PlanetProperty, thString },
    { "BASE.BUILDING",             ippBaseBuildFlag,     PlanetProperty, thBool },
    { "BASE.YESNO",                ippBaseFlag,          PlanetProperty, thBool },
    { "BUILD",                     ibpBuildHullName,     BaseProperty,   thString },
    { "BUILD.BEAM$",               ibpBuildBeam,         BaseProperty,   thInt },
    { "BUILD.BEAM.COUNT",          ibpBuildBeamCount,    BaseProperty,   thInt },
    { "BUILD.ENGINE$",             ibpBuildEngine,       BaseProperty,   thInt },
    { "BUILD.HULL$",               ibpBuildHull,         BaseProperty,   thInt },
    { "BUILD.QPOS",                ibpBuildQueuePos,     BaseProperty,   thInt },
    { "BUILD.SHORT",               ibpBuildHullShort,    BaseProperty,   thInt },
    { "BUILD.TORP$",               ibpBuildTorp,         BaseProperty,   thInt },
    { "BUILD.TORP.COUNT",          ibpBuildTorpCount,    BaseProperty,   thInt },
    { "BUILD.YESNO",               ibpBuildFlag,         BaseProperty,   thBool },
    { "BUILDBASE",                 ipmBuildBase,         PlanetMethod,   thProcedure },
    { "BUILDBASEDEFENSE",          ipmBuildBaseDefense,  PlanetMethod,   thProcedure },
    { "BUILDBEAMS",                ipmBuildBeams,        PlanetMethod,   thProcedure },
    { "BUILDDEFENSE",              ipmBuildDefense,      PlanetMethod,   thProcedure },
    { "BUILDENGINES",              ipmBuildEngines,      PlanetMethod,   thProcedure },
    { "BUILDFACTORIES",            ipmBuildFactories,    PlanetMethod,   thProcedure },
    { "BUILDFIGHTERS",             ipmBuildFighters,     PlanetMethod,   thProcedure },
    { "BUILDHULLS",                ipmBuildHulls,        PlanetMethod,   thProcedure },
    { "BUILDLAUNCHERS",            ipmBuildLaunchers,    PlanetMethod,   thProcedure },
    { "BUILDMINES",                ipmBuildMines,        PlanetMethod,   thProcedure },
    { "BUILDSHIP",                 ipmBuildShip,         PlanetMethod,   thProcedure },
    { "BUILDTORPS",                ipmBuildTorps,        PlanetMethod,   thProcedure },
    { "CARGOTRANSFER",             ipmCargoTransfer,     PlanetMethod,   thProcedure },
    { "COLONISTS",                 ippColonists,         PlanetProperty, thInt },
    { "COLONISTS.CHANGE",          ippColonistChangeStr, PlanetProperty, thString },
    { "COLONISTS.CHANGE$",         ippColonistChange,    PlanetProperty, thInt },
    { "COLONISTS.HAPPY",           ippColonistHappyStr,  PlanetProperty, thString },
    { "COLONISTS.HAPPY$",          ippColonistHappy,     PlanetProperty, thInt },
    { "COLONISTS.SUPPORTED",       ippColonistSupported, PlanetProperty, thInt },
    { "COLONISTS.TAX",             ippColonistTax,       PlanetProperty, thInt },
    { "COLONISTS.TAX.INCOME",      ippColonistTaxIncome, PlanetProperty, thInt },
    { "DAMAGE",                    ibpBaseDamage,        BaseProperty,   thInt },
    { "DEFENSE",                   ippDefense,           PlanetProperty, thInt },
    { "DEFENSE.BASE",              ibpBaseDefense,       BaseProperty,   thInt },
    { "DEFENSE.BASE.MAX",          ibpBaseDefenseMax,    BaseProperty,   thInt },
    { "DEFENSE.BASE.WANT",         ippBaseDefenseWanted, PlanetProperty, thInt },
    { "DEFENSE.MAX",               ippDefenseMax,        PlanetProperty, thInt },
    { "DEFENSE.WANT",              ippDefenseWanted,     PlanetProperty, thInt },
    { "DENSITY.D",                 ippDensityD,          PlanetProperty, thInt },
    { "DENSITY.M",                 ippDensityM,          PlanetProperty, thInt },
    { "DENSITY.N",                 ippDensityN,          PlanetProperty, thInt },
    { "DENSITY.T",                 ippDensityT,          PlanetProperty, thInt },
    { "FACTORIES",                 ippFactories,         PlanetProperty, thInt },
    { "FACTORIES.MAX",             ippFactoriesMax,      PlanetProperty, thInt },
    { "FACTORIES.WANT",            ippFactoriesWanted,   PlanetProperty, thInt },
    { "FCODE",                     ippFCode,             PlanetProperty, thString },
    { "FIGHTERS",                  ibpBaseFighters,      BaseProperty,   thInt },
    { "FIGHTERS.MAX",              ibpBaseFightersMax,   BaseProperty,   thInt },
    { "FIXSHIP",                   ipmFixShip,           PlanetMethod,   thProcedure },
    { "GROUND.D",                  ippGroundD,           PlanetProperty, thInt },
    { "GROUND.M",                  ippGroundM,           PlanetProperty, thInt },
    { "GROUND.N",                  ippGroundN,           PlanetProperty, thInt },
    { "GROUND.T",                  ippGroundT,           PlanetProperty, thInt },
    { "ID",                        ippId,                PlanetProperty, thInt },
    { "INDUSTRY",                  ippIndustry,          PlanetProperty, thString },
    { "INDUSTRY$",                 ippIndustryCode,      PlanetProperty, thInt },
    { "LEVEL",                     ippLevel,             PlanetProperty, thInt },
    { "LOC.X",                     ippLocX,              PlanetProperty, thInt },
    { "LOC.Y",                     ippLocY,              PlanetProperty, thInt },
    { "MARK",                      ipmMark,              PlanetMethod,   thProcedure },
    { "MARKED",                    ippMarked,            PlanetProperty, thBool },
    { "MINED.D",                   ippMinedD,            PlanetProperty, thInt },
    { "MINED.M",                   ippMinedM,            PlanetProperty, thInt },
    { "MINED.N",                   ippMinedN,            PlanetProperty, thInt },
    { "MINED.STR",                 ippMinedStr,          PlanetProperty, thString},
    { "MINED.T",                   ippMinedT,            PlanetProperty, thInt },
    { "MINES",                     ippMines,             PlanetProperty, thInt },
    { "MINES.MAX",                 ippMinesMax,          PlanetProperty, thInt },
    { "MINES.WANT",                ippMinesWanted,       PlanetProperty, thInt },
    { "MISSION",                   ibpMissionName,       BaseProperty,   thString },
    { "MISSION$",                  ibpMission,           BaseProperty,   thInt },
    { "MONEY",                     ippMoney,             PlanetProperty, thInt },
    { "NAME",                      ippName,              PlanetProperty, thString },
    { "NATIVES",                   ippNatives,           PlanetProperty, thString },
    { "NATIVES.CHANGE",            ippNativeChangeStr,   PlanetProperty, thString },
    { "NATIVES.CHANGE$",           ippNativeChange,      PlanetProperty, thInt },
    { "NATIVES.GOV",               ippNativeGov,         PlanetProperty, thString },
    { "NATIVES.GOV$",              ippNativeGovCode,     PlanetProperty, thInt },
    { "NATIVES.HAPPY",             ippNativeHappyStr,    PlanetProperty, thString },
    { "NATIVES.HAPPY$",            ippNativeHappy,       PlanetProperty, thInt },
    { "NATIVES.RACE",              ippNativeRace,        PlanetProperty, thString },
    { "NATIVES.RACE$",             ippNativeRaceCode,    PlanetProperty, thInt },
    { "NATIVES.TAX",               ippNativeTax,         PlanetProperty, thInt },
    { "NATIVES.TAX.BASE",          ippNativeTaxBase,     PlanetProperty, thInt },
    { "NATIVES.TAX.INCOME",        ippNativeTaxIncome,   PlanetProperty, thInt },
    { "NATIVES.TAX.MAX",           ippNativeTaxMax,      PlanetProperty, thInt },
    { "ORBIT",                     ippOrbitingShips,     PlanetProperty, thInt },
    { "ORBIT.ENEMY",               ippOrbitingEnemies,   PlanetProperty, thInt },
    { "ORBIT.OWN",                 ippOrbitingOwn,       PlanetProperty, thInt },
    { "OWNER",                     iplShortName,         OwnerProperty,  thString },
    { "OWNER$",                    iplId,                OwnerProperty,  thInt },
    { "OWNER.ADJ",                 iplAdjName,           OwnerProperty,  thString },
    { "PLAYED",                    ippPlayed,            PlanetProperty, thBool },
    { "RECYCLESHIP",               ipmRecycleShip,       PlanetMethod,   thProcedure },
    { "SELLSUPPLIES",              ipmSellSupplies,      PlanetMethod,   thProcedure },
    { "SETCOLONISTTAX",            ipmSetColonistTax,    PlanetMethod,   thProcedure },
    { "SETCOMMENT",                ipmSetComment,        PlanetMethod,   thProcedure },
    { "SETFCODE",                  ipmSetFCode,          PlanetMethod,   thProcedure },
    { "SETMISSION",                ipmSetMission,        PlanetMethod,   thProcedure },
    { "SETNATIVETAX",              ipmSetNativeTax,      PlanetMethod,   thProcedure },
    { "SETTECH",                   ipmSetTech,           PlanetMethod,   thProcedure },
    { "SHIPYARD",                  ibpShipyardStr,       BaseProperty,   thString },
    { "SHIPYARD.ACTION",           ibpShipyardAction,    BaseProperty,   thString },
    { "SHIPYARD.ID",               ibpShipyardId,        BaseProperty,   thInt },
    { "SHIPYARD.NAME",             ibpShipyardName,      BaseProperty,   thString },
    { "STORAGE.AMMO",              ibpAmmoStorage,       BaseProperty,   thArray },
    { "STORAGE.BEAMS",             ibpBeamStorage,       BaseProperty,   thArray },
    { "STORAGE.ENGINES",           ibpEngineStorage,     BaseProperty,   thArray },
    { "STORAGE.HULLS",             ibpHullStorage,       BaseProperty,   thArray },
    { "STORAGE.LAUNCHERS",         ibpLauncherStorage,   BaseProperty,   thArray },
    { "SUPPLIES",                  ippSupplies,          PlanetProperty, thInt },
    { "TASK",                      ippTask,              PlanetProperty, thBool },
    { "TASK.BASE",                 ippTaskBase,          PlanetProperty, thBool },
    { "TECH.BEAM",                 ibpBeamTech,          BaseProperty,   thInt },
    { "TECH.ENGINE",               ibpEngineTech,        BaseProperty,   thInt },
    { "TECH.HULL",                 ibpHullTech,          BaseProperty,   thInt },
    { "TECH.TORPEDO",              ibpTorpedoTech,       BaseProperty,   thInt },
    { "TEMP",                      ippTempStr,           PlanetProperty, thString },
    { "TEMP$",                     ippTemp,              PlanetProperty, thInt },
    { "UNMARK",                    ipmUnmark,            PlanetMethod,   thProcedure },
};

static const IntObjectProcedureValue::call_t planet_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
    IFPlanetSetComment,         // 2
    IFBaseFixShip,              // 3
    IFBaseRecycleShip,          // 4
    IFPlanetBuildBase,          // 5
    IFPlanetAutoBuild,          // 6
    IFPlanetBuildDefense,       // 7
    IFPlanetBuildFactories,     // 8
    IFPlanetBuildMines,         // 9
    IFPlanetSetColonistTax,     // 10
    IFPlanetSetNativeTax,       // 11
    IFPlanetSetFCode,           // 12
    IFBaseSetMission,           // 13
    IFPlanetBuildBaseDefense,   // 14
    IFBaseSetTech,              // 15
    IFBaseBuildFighters,        // 16
    IFBaseBuildEngines,         // 17
    IFBaseBuildHulls,           // 18
    IFBaseBuildLaunchers,       // 19
    IFBaseBuildBeams,           // 20
    IFBaseBuildTorps,           // 21
    IFPlanetSellSupplies,       // 22
    IFBaseBuildShip,            // 23
    IFPlanetCargoTransfer,      // 24
    IFPlanetAutoTaxColonists,   // 25
    IFPlanetAutoTaxNatives,     // 26
};

IntPlanetContext::IntPlanetContext(int pid)
    : pid(pid)
{ }

IntPlanetContext::~IntPlanetContext()
{ }

static int32_t
lookupPlanetProperty(const string_t& name)
{
    static const int32_t nproperties = countof(planet_mapping);

    /* Check user-defined properties */
    IntVariableNames::index_t ix = planet_property_names.lookup(name);
    if (ix != planet_property_names.nil)
        return ix + nproperties;

    /* Check predefined properties */
    return lookupName(name, planet_mapping, nproperties);
}

int32_t
IntPlanetContext::lookup(const string_t& name)
{
    if (name.compare(0, 7, "PLANET.", 7) == 0) {
        return lookupPlanetProperty(string_t(name, 7));
    } else {
        return lookupPlanetProperty(name);
    }
}

void
IntPlanetContext::set(int32_t index, IntValue* value)
{
    static const int32_t nproperties = countof(planet_mapping);
    if (index < nproperties) {
        /* Builtin property */
        GPlanet& pl = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
        switch (planet_mapping[index].domain) {
         case PlanetProperty:
            setPlanetProperty(pl, IntPlanetProperty(planet_mapping[index].index), value);
            break;
         case BaseProperty:
            setBaseProperty(pl, IntBaseProperty(planet_mapping[index].index), value);
            break;
         default:
            throw IntError::notAssignable();
        }
    } else {
        /* User property */
        getPlanetProperties(pid).set(index - nproperties, value);
        getDisplayedTurn().getCurrentUniverse().getPlanet(pid).markDirty();
    }
}

IntValue*
IntPlanetContext::get(int32_t index)
{
    static const int32_t nproperties = countof(planet_mapping);
    if (index < nproperties) {
        /* Builtin property */
        GPlanet& pl = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
        switch (planet_mapping[index].domain) {
         case PlanetProperty:
            return getPlanetProperty(pl, IntPlanetProperty(planet_mapping[index].index));
         case BaseProperty:
            return getBaseProperty(pl, IntBaseProperty(planet_mapping[index].index));
         case OwnerProperty:
            if (pl.isExistant() && pl.isOwnerKnown())
                return getPlayerProperty(pl.getOwner(), IntPlayerProperty(planet_mapping[index].index));
            else
                return 0;
         case PlanetMethod:
            return new IntObjectProcedureValue(pl, planet_methods[planet_mapping[index].index]);
        }
        return 0;
    } else {
        /* User property */
        return IntValue::cloneOf(getPlanetProperties(pid)[index - nproperties]);
    }
}

bool
IntPlanetContext::next()
{
    int n = global_anyplanet_type.getNextNoWrap(pid, false);
    if (n != 0) {
        pid = n;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntPlanetContext::clone()
{
    return new IntPlanetContext(pid);
}

GObject*
IntPlanetContext::getObject()
{
    return &getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
}

string_t
IntPlanetContext::toString(bool /*readable*/)
{
    return format("Planet(%d)", pid);
}

void
IntPlanetContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Planet;
    sv.value = pid;
}

void
IntPlanetContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumNames(planet_property_names);
    acceptor.enumMapping(planet_mapping, countof(planet_mapping));
}

template<typename T>
static IntValue*
makeIntFromMaybe(T value)
{
    if (value.isKnown())
        return makeIntValue(value);
    else
        return 0;
}

static int
countShips(GPoint pt, GPlayerSet players)
{
    int n = 0;
    for (int sid = global_anyship_type.getNextNoWrap(0, false); sid != 0; sid = global_anyship_type.getNextNoWrap(sid, false)) {
        const GShip& sh = global_anyship_type.getObject(sid);
        if (pt == sh.getPos() && players.contains(sh.getOwner()))
            ++n;
    }
    return n;
}

IntValue*
getPlanetProperty(const GPlanet& pl, IntPlanetProperty ipp)
{
    TAutobuildSettings abs;

    /* For non-existant planets, we only know Name and Id */
    /* FIXME: is this needed? */
    if (!pl.isExistant() && ipp != ippId && ipp != ippName)
        return 0;

    switch (ipp) {
     case ippBaseDefenseWanted:
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_BaseDefense]);
     case ippBaseFlag:
        return makeBoolValue(pl.hasBase());
     case ippBaseBuildFlag:
        if (pl.hasPlanetData()) {
            return makeBoolValue(pl.isBuildingBase());
        } else {
            return 0;
        }
     case ippBaseStr:
        if (pl.hasBase())
            return makeStringValue("present");
        else if (pl.isBuildingBase())
            return makeStringValue("being built");
        else
            return makeStringValue("-");
     case ippColonistChange:
        return makeIntFromMaybe(pl.getColonistChange());
     case ippColonistChangeStr:
        {
            mn32_t change = pl.getColonistChange();
            if (change.isKnown())
                return makeStringValue(getHappyChangeName(change));
            else
                return 0;
        }
     case ippColonistHappy:
        return makeIntFromMaybe(pl.getColonistHappiness());
     case ippColonistHappyStr:
        if (pl.getColonistHappiness().isKnown())
            return makeStringValue(getHappinessName(pl.getColonistHappiness()));
        else
            return 0;
     case ippColonistSupported:
        return makeIntFromMaybe(pl.getMaxSupportedColonists());
     case ippColonistTax:
        return makeIntFromMaybe(pl.getColonistTax());
     case ippColonistTaxIncome:
        if (pl.isOwnerKnown() && pl.getOwner() > 0 && pl.getColonists().isKnown() && pl.getColonistTax().isKnown()) {
            return makeIntValue(pl.getColonistDue(pl.getColonistTax()));
        } else {
            return 0;
        }
     case ippColonists:
        return makeIntFromMaybe(pl.getColonists());
     case ippDefense:
        return makeIntFromMaybe(pl.getStructures(str_Defense));
     case ippDefenseMax:
        return makeIntFromMaybe(pl.getMaxStructures(str_Defense));
     case ippDefenseWanted:
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Defense]);
     case ippDensityD:
        return makeIntFromMaybe(pl.getOreDensity(el_Duranium));
     case ippDensityM:
        return makeIntFromMaybe(pl.getOreDensity(el_Molybdenum));
     case ippDensityN:
        return makeIntFromMaybe(pl.getOreDensity(el_Neutronium));
     case ippDensityT:
        return makeIntFromMaybe(pl.getOreDensity(el_Tritanium));
     case ippFCode:
        if (pl.isFCodeKnown())
            return makeStringValue(pl.getFCode());
        else
            return 0;
     case ippFactories:
        return makeIntFromMaybe(pl.getStructures(str_Factories));
     case ippFactoriesMax:
        return makeIntFromMaybe(pl.getMaxStructures(str_Factories));
     case ippFactoriesWanted:
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Factories]);
     case ippGroundD:
        return makeIntFromMaybe(pl.getOreGround(el_Duranium));
     case ippGroundM:
        return makeIntFromMaybe(pl.getOreGround(el_Molybdenum));
     case ippGroundN:
        return makeIntFromMaybe(pl.getOreGround(el_Neutronium));
     case ippGroundT:
        return makeIntFromMaybe(pl.getOreGround(el_Tritanium));
     case ippId:
        return makeIntValue(pl.getId());
     case ippIndustry:
        if (pl.getIndustryLevel().isKnown())
            return makeStringValue(getIndustryLevelName(pl.getIndustryLevel()));
        else
            return 0;
     case ippIndustryCode:
        return makeIntFromMaybe(pl.getIndustryLevel());
     case ippLevel:
        {
            int16_t level = pl.unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpLevel));
            if (level >= 0)
                return makeIntValue(level);
            else
                return 0;
        }
     case ippLocX:
        return makeIntValue(pl.getPos().x);
     case ippLocY:
        return makeIntValue(pl.getPos().y);
     case ippMarked:
        return makeBoolValue(pl.isMarked());
     case ippMinedD:
        if (pl.isCargoAmountKnown(el_Duranium))
            return makeIntValue(pl.getCargo(el_Duranium));
        else
            return 0;
     case ippMinedM:
        if (pl.isCargoAmountKnown(el_Molybdenum))
            return makeIntValue(pl.getCargo(el_Molybdenum));
        else
            return 0;
     case ippMinedN:
        if (pl.isCargoAmountKnown(el_Neutronium))
            return makeIntValue(pl.getCargo(el_Neutronium));
        else
            return 0;
     case ippMinedStr:
        {
            GCargoSpec sp;
            bool haveAny = false;
            for (GCargoType el = el_Neutronium; el <= el_Molybdenum; ++el) {
                if (pl.isCargoAmountKnown(el)) {
                    sp.set(el, pl.getCargo(el));
                    haveAny = true;
                }
            }
            if (haveAny)
                return makeStringValue(sp.toCargoSpecString());
            else
                return 0;
        }
     case ippMinedT:
        if (pl.isCargoAmountKnown(el_Tritanium))
            return makeIntValue(pl.getCargo(el_Tritanium));
        else
            return 0;
     case ippMines:
        return makeIntFromMaybe(pl.getStructures(str_Mines));
     case ippMinesMax:
        return makeIntFromMaybe(pl.getMaxStructures(str_Mines));
     case ippMinesWanted:
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Mines]);
     case ippMoney:
        if (pl.isCargoAmountKnown(el_Money))
            return makeIntValue(pl.getCargo(el_Money));
        else
            return 0;
     case ippName:
        return makeStringValue(pl.getName());
     case ippNativeChange:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeChange());
        else
            return 0;
     case ippNativeChangeStr:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeChange().isKnown())
            return makeStringValue(getHappyChangeName(pl.getNativeChange()));
        else
            return 0;
     case ippNativeGov:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeGovernment().isKnown())
            return makeStringValue(getNativeGovernmentName(pl.getNativeGovernment()));
        else
            return 0;
     case ippNativeGovCode:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeGovernment());
        else
            return 0;
     case ippNativeHappy:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeHappiness());
        else
            return 0;
     case ippNativeHappyStr:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeHappiness().isKnown())
            return makeStringValue(getHappinessName(pl.getNativeHappiness()));
        else
            return 0;
     case ippNativeRace:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeRace().isKnown())
            return makeStringValue(getNativeRaceName(pl.getNativeRace()));
        else
            return 0;
     case ippNativeRaceCode:
        if (pl.getNatives().isKnown())
            return makeIntFromMaybe(pl.getNativeRace());
        else
            return 0;
     case ippNativeTax:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeTax());
        else
            return 0;
     case ippNativeTaxBase:
        return makeIntFromMaybe(pl.getNativeBaseTax(0));
     case ippNativeTaxMax:
        return makeIntFromMaybe(pl.getNativeBaseTax(-30));
     case ippNativeTaxIncome:
        if (pl.getNatives().isKnown()
            && pl.getNativeRace().isKnown()
            && pl.getNativeGovernment().isKnown()
            && pl.getNativeTax().isKnown())
        {
            if (pl.getNatives() > 0 && pl.isOwnerKnown() && pl.getOwner() > 0) {
                return makeIntValue(pl.getNativeDue(pl.getNativeTax()));
            } else {
                // No natives, or no owner: income is 0
                return makeIntValue(0);
            }
        } else {
            return 0;
        }
     case ippNatives:
        return makeIntFromMaybe(pl.getNatives());
     case ippOrbitingEnemies:
        return makeIntValue(countShips(pl.getPos(), GPlayerSet::allUpTo(NUM_OWNERS) - getPlayerId()));
     case ippOrbitingOwn:
        return makeIntValue(countShips(pl.getPos(), GPlayerSet(getPlayerId())));
     case ippOrbitingShips:
        return makeIntValue(countShips(pl.getPos(), GPlayerSet::allUpTo(NUM_OWNERS)));
     case ippPlayed:
        return makeBoolValue(pl.isPlayed());
     case ippSupplies:
        if (pl.isCargoAmountKnown(el_Supplies))
            return makeIntValue(pl.getCargo(el_Supplies));
        else
            return 0;
     case ippTask:
        return makeBoolValue(getProcessByObject(&pl, IntExecutionContext::pkPlanetTask) != 0);
     case ippTaskBase:
        return makeBoolValue(getProcessByObject(&pl, IntExecutionContext::pkBaseTask) != 0);
     case ippTemp:
        return makeIntFromMaybe(pl.getTemperature());
     case ippTempStr:
        if (pl.getTemperature().isKnown())
            return makeStringValue(getTemperatureName(pl.getTemperature()));
        else
            return 0;
    }
    return 0;
}

void
setPlanetProperty(GPlanet& pl, IntPlanetProperty ipp, IntValue* value)
{
    // We cannot do anything on nonexistant planets
    if (!pl.isExistant())
        throw IntError::notAssignable();

    // We cannot assign to anything other than auto-build goals on non-played planets
    if (!pl.isPlayed() && (ipp != ippMinesWanted && ipp != ippFactoriesWanted && ipp != ippDefenseWanted && ipp != ippBaseDefenseWanted))
        throw IntError::notAssignable();

    int32_t iv;
    switch (ipp) {
     case ippMinesWanted:
        if (checkIntArg(iv, value, 0, 1000))
            pl.setAutobuildGoal(str_Mines, iv);
        break;
     case ippFactoriesWanted:
        if (checkIntArg(iv, value, 0, 1000))
            pl.setAutobuildGoal(str_Factories, iv);
        break;
     case ippDefenseWanted:
        if (checkIntArg(iv, value, 0, 1000))
            pl.setAutobuildGoal(str_Defense, iv);
        break;
     case ippBaseDefenseWanted:
        if (checkIntArg(iv, value, 0, 1000))
            pl.setAutobuildGoal(str_BaseDefense, iv);
        break;
     case ippColonistTax:
        if (checkIntArg(iv, value, 0, 100))
            pl.setColonistTax(mp16_t(iv));
        break;
     case ippFCode:
        {
            string_t sv;
            if (checkStringArg(sv, value))
                pl.setFCode(sv);
            break;
        }
     case ippNativeTax:
        if (checkIntArg(iv, value, 0, 100)) {
            if (pl.getNativeRace() <= 0)
                throw IntError::notAssignable();
            else
                pl.setNativeTax(mp16_t(iv));
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

void
callPlanetMethod(GPlanet& pl, IntPlanetMethod ipm, IntExecutionContext& exc, IntArgBlock& args)
{
    planet_methods[ipm](pl, exc, args);
}
